import functools
import typing as T
from pathlib import Path

from PyQt5 import QtCore, QtGui, QtWidgets

from .color_model import ColorModel
from .color_preview import ColorPreview
from .color_ring import ColorRing
from .color_sliders import (
    AlphaColorControl,
    BlueColorControl,
    GreenColorControl,
    HueColorControl,
    RedColorControl,
    SaturationColorControl,
    ValueColorControl,
)
from .color_square import ColorSquare, ColorSquareStyle
from .screen_color_picker import ScreenColorPicker


class ButtonStrip(QtWidgets.QDialogButtonBox):
    reset = QtCore.pyqtSignal()
    pick = QtCore.pyqtSignal()

    def __init__(self, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent)
        pick_button = QtWidgets.QPushButton("Pick screen color")
        self.addButton(self.Reset)
        self.addButton(pick_button, self.ActionRole)
        self.addButton(self.Ok)
        self.addButton(self.Cancel)
        self.button(self.Reset).clicked.connect(self.reset.emit)
        pick_button.clicked.connect(self.pick.emit)


class ClickableLabel(QtWidgets.QLabel):
    pressed = QtCore.pyqtSignal()
    released = QtCore.pyqtSignal()

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        if event.button() != QtCore.Qt.LeftButton:
            event.ignore()
        event.accept()
        self.pressed.emit()

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        if event.button() != QtCore.Qt.LeftButton:
            event.ignore()
        event.accept()
        self.released.emit()


class SlidersControl(QtWidgets.QWidget):
    def __init__(
        self,
        parent: QtWidgets.QWidget,
        model: ColorModel,
        alpha_grid: QtGui.QPixmap,
    ) -> None:
        super().__init__(parent)
        layout = QtWidgets.QGridLayout(self)
        layout.setSpacing(8)
        layout.setContentsMargins(0, 0, 0, 0)

        def mouse_down(radio_button: QtWidgets.QRadioButton) -> None:
            radio_button.setDown(True)

        def mouse_up(radio_button: QtWidgets.QRadioButton) -> None:
            radio_button.setDown(False)
            radio_button.setChecked(True)
            radio_button.clicked.emit()

        self.radio_buttons = {
            square_style: QtWidgets.QRadioButton(self)
            for square_style in ColorSquareStyle
        }
        self.labels = {
            square_style: ClickableLabel(square_style.name.title() + ":", self)
            for square_style in ColorSquareStyle
        }

        for label, radio_button in zip(
            self.labels.values(), self.radio_buttons.values()
        ):
            label.pressed.connect(functools.partial(mouse_down, radio_button))
            label.released.connect(functools.partial(mouse_up, radio_button))

        layout.addWidget(self.radio_buttons[ColorSquareStyle.Hue], 0, 0)
        layout.addWidget(self.radio_buttons[ColorSquareStyle.Saturation], 1, 0)
        layout.addWidget(self.radio_buttons[ColorSquareStyle.Value], 2, 0)
        layout.addWidget(self.labels[ColorSquareStyle.Hue], 0, 1)
        layout.addWidget(self.labels[ColorSquareStyle.Saturation], 1, 1)
        layout.addWidget(self.labels[ColorSquareStyle.Value], 2, 1)
        layout.addWidget(HueColorControl(self, model, alpha_grid), 0, 2)
        layout.addWidget(SaturationColorControl(self, model, alpha_grid), 1, 2)
        layout.addWidget(ValueColorControl(self, model, alpha_grid), 2, 2)

        layout.addWidget(QtWidgets.QFrame(self), 3, 1, 1, 2)

        layout.addWidget(self.radio_buttons[ColorSquareStyle.Red], 4, 0)
        layout.addWidget(self.radio_buttons[ColorSquareStyle.Green], 5, 0)
        layout.addWidget(self.radio_buttons[ColorSquareStyle.Blue], 6, 0)
        layout.addWidget(self.labels[ColorSquareStyle.Red], 4, 1)
        layout.addWidget(self.labels[ColorSquareStyle.Green], 5, 1)
        layout.addWidget(self.labels[ColorSquareStyle.Blue], 6, 1)
        layout.addWidget(RedColorControl(self, model, alpha_grid), 4, 2)
        layout.addWidget(GreenColorControl(self, model, alpha_grid), 5, 2)
        layout.addWidget(BlueColorControl(self, model, alpha_grid), 6, 2)

        self.alpha_widgets = [
            QtWidgets.QFrame(self),
            QtWidgets.QLabel("Opacity:", self),
            AlphaColorControl(self, model, alpha_grid),
        ]
        layout.addWidget(self.alpha_widgets[0], 7, 0, 1, 2)
        layout.addWidget(self.alpha_widgets[1], 8, 1)
        layout.addWidget(self.alpha_widgets[2], 8, 2)


class QColorDialog(QtWidgets.QDialog):
    ColorDialogOption = QtWidgets.QColorDialog.ColorDialogOption
    ColorDialogOptions = QtWidgets.QColorDialog.ColorDialogOptions
    ShowAlphaChannel = QtWidgets.QColorDialog.ShowAlphaChannel
    NoButtons = QtWidgets.QColorDialog.NoButtons
    DontUseNativeDialog = QtWidgets.QColorDialog.DontUseNativeDialog

    colorSelected = QtCore.pyqtSignal(QtGui.QColor)
    currentColorChanged = QtCore.pyqtSignal(QtGui.QColor)

    def __init__(
        self,
        initial: T.Optional[QtGui.QColor] = None,
        parent: QtWidgets.QWidget = None,
    ) -> None:
        super().__init__(parent)
        initial = (
            initial if initial is not None else QtGui.QColor(255, 255, 255)
        )

        self._use_square_view = False
        self._initial = initial
        self._model = ColorModel(initial)
        self._screen_color_picker = ScreenColorPicker(self._model, self)
        self._options = (
            QtWidgets.QColorDialog.ColorDialogOptions() | self.ShowAlphaChannel
        )
        self._alpha_grid = QtGui.QPixmap(
            str(Path(__file__).parent / "grid.png")
        )

        self._color_square = ColorSquare(self, self._model)
        self._color_ring = ColorRing(self, self._model)
        self._color_preview = ColorPreview(
            self, initial, self._model, self._alpha_grid
        )
        self._sliders = SlidersControl(self, self._model, self._alpha_grid)
        self._strip = ButtonStrip(self)

        left_layout = QtWidgets.QVBoxLayout()
        left_layout.setSpacing(16)
        left_layout.setContentsMargins(0, 0, 16, 0)
        left_layout.addWidget(self._color_square)
        left_layout.addWidget(self._color_ring)
        left_layout.addWidget(self._color_preview)

        right_layout = QtWidgets.QVBoxLayout()
        right_layout.setSpacing(32)
        right_layout.addWidget(self._sliders)
        right_layout.addWidget(self._strip)

        root_layout = QtWidgets.QHBoxLayout(self)
        root_layout.addLayout(left_layout)
        root_layout.addLayout(right_layout)

        self._sliders.radio_buttons[
            self._color_square.square_style
        ].setChecked(True)

        for square_style, radio_button in self._sliders.radio_buttons.items():
            radio_button.clicked.connect(
                functools.partial(
                    self._color_square.set_square_style, square_style
                )
            )
        self._strip.accepted.connect(self.accept)
        self._strip.rejected.connect(self.reject)
        self._strip.reset.connect(self.reset)
        self._strip.pick.connect(self._screen_color_picker.pick_screen_color)
        self._model.changed.connect(self._emit_signal)

        self.setWindowTitle("Select color")
        self.layout().setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.show()

    def options(self) -> QtWidgets.QColorDialog.ColorDialogOptions:
        return self._options

    def testOption(
        self, option: QtWidgets.QColorDialog.ColorDialogOption
    ) -> bool:
        return self._options & option

    def setOption(
        self, option: QtWidgets.QColorDialog.ColorDialogOption, on: bool = True
    ) -> None:
        if on:
            self._options |= option
        else:
            self._options &= ~option
        self._update_options()

    def setOptions(
        self, options: QtWidgets.QColorDialog.ColorDialogOptions
    ) -> None:
        self._options = options
        self._update_options()

    def setCurrentColor(self, color: QtGui.QColor) -> None:
        self._model.color = color

    def reset(self) -> None:
        self._model.color = self._initial

    def currentColor(self) -> QtGui.QColor:
        return self._model.color

    def selectedColor(self) -> QtGui.QColor:
        return self._model.color

    def alphaGrid(self) -> QtGui.QPixmap:
        return self._alpha_grid

    def setAlphaGrid(self, alpha_grid: QtGui.QPixmap) -> None:
        self._alpha_grid.convertFromImage(alpha_grid.toImage())
        self.update()

    def useSquareView(self) -> bool:
        return self._use_square_view

    def setUseSquareView(self, use_square_view: bool) -> None:
        self._use_square_view = use_square_view
        self._update_options()

    @staticmethod
    def getColor(
        initial: T.Optional[QtGui.QColor] = None,
        parent: T.Optional[QtWidgets.QWidget] = None,
        title: T.Optional[str] = None,
        options: T.Optional[QtWidgets.QColorDialog.ColorDialogOptions] = None,
        use_square_view: T.Optional[bool] = False,
        alpha_grid: T.Optional[QtGui.QPixmap] = None,
    ) -> QtGui.QColor:
        dialog = QColorDialog(initial, parent)
        if title is not None:
            dialog.setWindowTitle(title)
        if options is not None:
            dialog.setOptions(options)
        if alpha_grid is not None:
            dialog.setAlphaGrid(alpha_grid)
        if use_square_view is not None:
            dialog.setUseSquareView(use_square_view)
        ret = dialog.exec_()
        if ret == QtWidgets.QDialog.Accepted:
            return dialog.selectedColor()
        return QtGui.QColor()

    def _update_options(self) -> None:
        for widget in self._sliders.alpha_widgets:
            widget.setVisible(self._options & self.ShowAlphaChannel)
        self._color_square.setVisible(self._use_square_view)
        self._color_ring.setVisible(not self._use_square_view)
        for widget in self._sliders.radio_buttons.values():
            widget.setVisible(self._use_square_view)
        self._strip.setVisible(not self._options & self.NoButtons)

    def _emit_signal(self) -> None:
        self.colorSelected.emit(self.selectedColor())
        self.currentColorChanged.emit(self.currentColor())



import typing as T

from PyQt5 import QtCore, QtGui, QtWidgets

from .color_model import ColorModel
from .util import is_imprecise_click, is_precise_click


class ColorSlider(QtWidgets.QAbstractSlider):
    _thumb_size = 8

    def __init__(
        self,
        parent: QtWidgets.QWidget,
        gradient_decorator: T.Callable[[QtGui.QLinearGradient], None],
        alpha_grid: QtGui.QPixmap,
        value: int,
    ) -> None:
        super().__init__(
            parent,
            minimum=0,
            maximum=255,
            orientation=QtCore.Qt.Horizontal,
            value=value,
        )
        self._gradient_decorator = gradient_decorator
        self._alpha_grid = alpha_grid
        self._pressed_control = QtWidgets.QStyle.SC_None

        self.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Maximum
        )

    def setValue(self, value: float) -> None:
        super().setValue(value)
        self.update()

    def sizeHint(self) -> T.Tuple[int, int]:
        return QtCore.QSize(300, 25)

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        groove_rect = self._groove_rect
        handle_rect = self._handle_rect

        gradient = QtGui.QLinearGradient(
            groove_rect.topLeft(), groove_rect.topRight()
        )
        self._gradient_decorator(gradient)

        painter = QtGui.QPainter(self)
        painter.drawTiledPixmap(groove_rect, self._alpha_grid)
        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(gradient)
        painter.drawRect(groove_rect)

        opt = QtWidgets.QStyleOptionFrame()
        opt.rect = groove_rect
        opt.state = QtWidgets.QStyle.State_Sunken
        opt.lineWidth = 1
        opt.frameShape = QtWidgets.QFrame.Panel
        self.style().drawControl(
            QtWidgets.QStyle.CE_ShapedFrame, opt, painter, self
        )

        opt = QtWidgets.QStyleOptionButton()
        opt.state = (
            QtWidgets.QStyle.State_Active | QtWidgets.QStyle.State_Enabled
        )
        if self.isSliderDown():
            opt.state |= QtWidgets.QStyle.State_Sunken
        opt.rect = handle_rect
        self.style().drawControl(
            QtWidgets.QStyle.CE_PushButton, opt, painter, self
        )

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        if is_precise_click(event.button(), self.style()):
            event.accept()
            self.setSliderPosition(self._val_from_point(event.pos()))
            self.triggerAction(QtWidgets.QSlider.SliderMove)
            self._pressed_control = QtWidgets.QStyle.SC_SliderHandle
            self.update()

        elif is_imprecise_click(event.button(), self.style()):
            event.accept()

            self._pressed_control = (
                QtWidgets.QStyle.SC_SliderHandle
                if event.pos() in self._handle_rect
                else QtWidgets.QStyle.SC_SliderGroove
            )

            action = self.SliderNoAction
            if self._pressed_control == QtWidgets.QStyle.SC_SliderGroove:
                press_value = self._val_from_point(event.pos())
                if press_value > self.value():
                    action = self.SliderPageStepAdd
                elif press_value < self.value():
                    action = self.SliderPageStepSub
                if action:
                    self.triggerAction(action)
                    self.setRepeatAction(action)

        else:
            event.ignore()
            return

        if self._pressed_control == QtWidgets.QStyle.SC_SliderHandle:
            self.setRepeatAction(self.SliderNoAction)
            self.update()
            self.setSliderDown(True)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        if self._pressed_control != QtWidgets.QStyle.SC_SliderHandle:
            event.ignore()
            return
        event.accept()
        self.setSliderPosition(self._val_from_point(event.pos()))

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        if (
            self._pressed_control == QtWidgets.QStyle.SC_None
            or event.buttons()
        ):
            event.ignore()
            return
        event.accept()
        old_pressed = QtWidgets.QStyle.SubControl(self._pressed_control)
        self._pressed_control = QtWidgets.QStyle.SC_None
        self.setRepeatAction(self.SliderNoAction)
        if old_pressed == QtWidgets.QStyle.SC_SliderHandle:
            self.setSliderDown(False)
        self.update()

    @property
    def _groove_rect(self) -> QtCore.QRect:
        return QtCore.QRect(
            self._thumb_size,
            3,
            self.width() - self._thumb_size * 2,
            self.height() - 6,
        )

    @property
    def _handle_rect(self) -> QtCore.QRect:
        x = (self.value() - self.minimum()) / (self.maximum() - self.minimum())
        x = x * (self.width() - self._thumb_size * 2)
        return QtCore.QRect(x, 0, self._thumb_size * 2, self.height())

    def _val_from_point(self, pos: QtCore.QPoint) -> int:
        center = self._handle_rect.center() - self._handle_rect.topLeft()
        x = (pos - center).x()
        return int(
            self.minimum()
            + x * (self.maximum() - self.minimum()) / self._groove_rect.width()
        )


class BaseColorControl(QtWidgets.QWidget):
    def __init__(
        self,
        parent: QtWidgets.QWidget,
        model: ColorModel,
        alpha_grid: QtGui.QPixmap,
    ) -> None:
        super().__init__(parent)
        self._model = model

        self._slider = ColorSlider(
            self,
            lambda gradient: self._decorate_gradient(
                gradient, self._model.color
            ),
            alpha_grid,
            value=int(self._get_value(self._model) * 255),
        )
        self._up_down = QtWidgets.QSpinBox(
            self,
            minimum=0,
            maximum=255,
            value=int(self._get_value(self._model) * 255),
        )

        layout = QtWidgets.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._slider)
        layout.addWidget(self._up_down)

        self._slider.valueChanged.connect(self._slider_changed)
        self._up_down.valueChanged.connect(self._up_down_changed)
        self._model.changed.connect(self._model_changed)

    def _get_value(self, model: ColorModel) -> float:
        raise NotImplementedError("not implemented")

    def _set_value(self, model: ColorModel, value: float) -> None:
        raise NotImplementedError("not implemented")

    def _decorate_gradient(
        self, gradient: QtGui.QLinearGradient, color: QtGui.QColor
    ) -> None:
        tmp_model = ColorModel(self._model.color)
        tmp_model.a = 1.0
        self._set_value(tmp_model, 0.0)
        gradient.setColorAt(0, tmp_model.color)
        self._set_value(tmp_model, 1.0)
        gradient.setColorAt(1, tmp_model.color)

    def _model_changed(self) -> None:
        self._slider.setValue(self._get_value(self._model) * 255)
        self._up_down.setValue(self._get_value(self._model) * 255)

    def _slider_changed(self) -> None:
        self._set_value(self._model, self._slider.value() / 255.0)

    def _up_down_changed(self) -> None:
        self._set_value(self._model, self._up_down.value() / 255.0)


class HueColorControl(BaseColorControl):
    def _get_value(self, model: ColorModel) -> float:
        return model.h

    def _set_value(self, model: ColorModel, value: float) -> None:
        model.h = value

    def _decorate_gradient(
        self, gradient: QtGui.QLinearGradient, color: QtGui.QColor
    ) -> None:
        gradient.setColorAt(0 / 6, QtGui.QColor(255, 0, 0, 255))
        gradient.setColorAt(1 / 6, QtGui.QColor(255, 255, 0, 255))
        gradient.setColorAt(2 / 6, QtGui.QColor(0, 255, 0, 255))
        gradient.setColorAt(3 / 6, QtGui.QColor(0, 255, 255, 255))
        gradient.setColorAt(4 / 6, QtGui.QColor(0, 0, 255, 255))
        gradient.setColorAt(5 / 6, QtGui.QColor(255, 0, 255, 255))
        gradient.setColorAt(6 / 6, QtGui.QColor(255, 0, 0, 255))


class SaturationColorControl(BaseColorControl):
    def _get_value(self, model: ColorModel) -> float:
        return model.s

    def _set_value(self, model: ColorModel, value: float) -> None:
        model.s = value


class ValueColorControl(BaseColorControl):
    def _get_value(self, model: ColorModel) -> float:
        return model.v

    def _set_value(self, model: ColorModel, value: float) -> None:
        model.v = value


class RedColorControl(BaseColorControl):
    def _get_value(self, model: ColorModel) -> float:
        return model.r

    def _set_value(self, model: ColorModel, value: float) -> None:
        model.r = value


class GreenColorControl(BaseColorControl):
    def _get_value(self, model: ColorModel) -> float:
        return model.g

    def _set_value(self, model: ColorModel, value: float) -> None:
        model.g = value


class BlueColorControl(BaseColorControl):
    def _get_value(self, model: ColorModel) -> float:
        return model.b

    def _set_value(self, model: ColorModel, value: float) -> None:
        model.b = value


class AlphaColorControl(BaseColorControl):
    def _get_value(self, model: ColorModel) -> float:
        return model.a

    def _set_value(self, model: ColorModel, value: float) -> None:
        model.a = value

    def _decorate_gradient(
        self, gradient: QtGui.QLinearGradient, color: QtGui.QColor
    ) -> None:
        gradient.setColorAt(
            0, QtGui.QColor(color.red(), color.green(), color.blue(), 0)
        )
        gradient.setColorAt(
            1, QtGui.QColor(color.red(), color.green(), color.blue(), 255)
        )



import typing as T

from PyQt5 import QtCore


def point_in_triangle(
    point: QtCore.QPoint,
    triangle_points: T.Tuple[QtCore.QPoint, QtCore.QPoint, QtCore.QPoint],
) -> bool:
    p = point
    p1, p2, p3 = triangle_points
    area = 0.5 * (
        -p2.y() * p3.x()
        + p1.y() * (-p2.x() + p3.x())
        + p1.x() * (p2.y() - p3.y())
        + p2.x() * p3.y()
    )
    s = (
        1
        / (2 * area)
        * (
            p1.y() * p3.x()
            - p1.x() * p3.y()
            + (p3.y() - p1.y()) * p.x()
            + (p1.x() - p3.x()) * p.y()
        )
    )
    t = (
        1
        / (2 * area)
        * (
            p1.x() * p2.y()
            - p1.y() * p2.x()
            + (p1.y() - p2.y()) * p.x()
            + (p2.x() - p1.x()) * p.y()
        )
    )
    return s > 0 and t > 0 and 1 - s - t > 0


def point_in_ring(
    point: QtCore.QPoint, inner_radius: float, outer_radius: float
) -> bool:
    dist = (point.y() * point.y() + point.x() * point.x()) ** 0.5
    return inner_radius <= dist <= outer_radius



from PyQt5 import QtCore, QtGui, QtWidgets


def is_precise_click(
    button: QtCore.Qt.MouseButton, style: QtWidgets.QStyle
) -> bool:
    return (
        button & style.styleHint(QtWidgets.QStyle.SH_Slider_AbsoluteSetButtons)
    ) == button


def is_imprecise_click(
    button: QtCore.Qt.MouseButton, style: QtWidgets.QStyle
) -> bool:
    return (
        button & style.styleHint(QtWidgets.QStyle.SH_Slider_PageSetButtons)
    ) == button


def blend(c1: QtGui.QColor, c2: QtGui.QColor, ratio: float) -> QtGui.QColor:
    return QtGui.QColor.fromRgbF(
        c1.redF() * (1 - ratio) + c2.redF() * ratio,
        c1.greenF() * (1 - ratio) + c2.greenF() * ratio,
        c1.blueF() * (1 - ratio) + c2.blueF() * ratio,
    )


def clamp(val: float, min_val: float, max_val: float) -> float:
    return max(min_val, min(max_val, val))



from PyQt5 import QtCore, QtGui, QtWidgets

from .color_model import ColorModel


class ColorPickingEventFilter(QtCore.QObject):
    def __init__(self, parent: QtCore.QObject) -> None:
        super().__init__(parent)
        self._parent = parent

    def eventFilter(
        self, _object: QtCore.QObject, event: QtCore.QEvent
    ) -> bool:
        if event.type() == QtCore.QEvent.MouseMove:
            return self._parent.handle_mouse_move(event)
        elif event.type() == QtCore.QEvent.MouseButtonRelease:
            return self._parent.handle_mouse_button_release(event)
        elif event.type() == QtCore.QEvent.KeyPress:
            return self._parent.handle_key_press(event)
        return False


class ScreenColorPicker(QtCore.QObject):
    def __init__(self, model: ColorModel, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent)
        self._model = model
        self._parent = parent
        self._dummy_transparent_window = QtGui.QWindow()
        self._dummy_transparent_window.resize(1, 1)
        self._dummy_transparent_window.setFlags(
            QtCore.Qt.Tool | QtCore.Qt.FramelessWindowHint
        )
        self._color_picking_event_filter = ColorPickingEventFilter(self)
        self._old_color = QtGui.QColor()

    def pick_screen_color(self) -> None:
        self._old_color = self._model.color
        self._dummy_transparent_window.show()
        self._parent.installEventFilter(self._color_picking_event_filter)
        self._parent.grabMouse(QtCore.Qt.CrossCursor)
        self._parent.grabKeyboard()
        self._parent.setMouseTracking(True)

    def _release(self) -> None:
        self._dummy_transparent_window.setVisible(False)
        self._parent.removeEventFilter(self._color_picking_event_filter)
        self._parent.releaseMouse()
        self._parent.releaseKeyboard()
        self._parent.setMouseTracking(False)

    def handle_mouse_move(self, event: QtGui.QMouseEvent) -> bool:
        self._model.color = self._grab_screen_color(event.globalPos())
        return True

    def handle_mouse_button_release(self, event: QtGui.QMouseEvent) -> bool:
        self._model.color = self._grab_screen_color(event.globalPos())
        self._release()
        return True

    def handle_key_press(self, event: QtGui.QKeyEvent) -> bool:
        if event.matches(QtGui.QKeySequence.Cancel):
            self._release()
            self._model.color = self._old_color
        elif event.key() in {QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter}:
            self._model.color = self._grab_screen_color(QtGui.QCursor.pos())
            self._release()
        event.accept()
        return True

    def _grab_screen_color(self, point: QtCore.QPoint) -> QtGui.QColor:
        desktop = QtWidgets.QApplication.desktop()
        pixmap = QtWidgets.QApplication.primaryScreen().grabWindow(
            desktop.winId(), point.x(), point.y(), 1, 1
        )
        return pixmap.toImage().pixelColor(0, 0)



import math
import typing as T

from PyQt5 import QtCore, QtGui


class ColorModel(QtCore.QObject):
    changed = QtCore.pyqtSignal()

    def __init__(self, color: QtGui.QColor) -> None:
        super().__init__()
        self._h: float = color.hueF()
        self._s: float = color.saturationF()
        self._v: float = color.valueF()
        self._r: float = color.redF()
        self._g: float = color.greenF()
        self._b: float = color.blueF()
        self._a: float = color.alphaF()
        self._syncing = False
        self._color = QtGui.QColor(color)

    @property
    def color(self) -> QtGui.QColor:
        return self._color

    @color.setter
    def color(self, color: QtGui.QColor) -> None:
        self._r = color.redF()
        self._g = color.greenF()
        self._b = color.blueF()
        self._sync_to_hsv()
        self.a = color.alphaF()

    def _sync_to_rgb(self) -> None:
        if self._syncing:
            return
        self._syncing = True
        self._color = QtGui.QColor.fromHsvF(self.h, self.s, self.v, self.a)
        self.r = self._color.redF()
        self.g = self._color.greenF()
        self.b = self._color.blueF()
        self._syncing = False

    def _sync_to_hsv(self) -> None:
        if self._syncing:
            return
        self._syncing = True
        self._color = QtGui.QColor.fromRgbF(self.r, self.g, self.b, self.a)
        self.h = self._color.hueF()
        self.s = self._color.saturationF()
        self.v = self._color.valueF()
        self._syncing = False

    @property
    def a(self) -> float:
        return self._a

    @a.setter
    def a(self, a: float) -> None:
        a = max(0.0, min(1.0, a))
        if a != self._a:
            self._a = a
            self.changed.emit()
            self._color.setAlphaF(a)

    @property
    def h(self) -> float:
        return self._h

    @h.setter
    def h(self, h: float) -> None:
        h = max(0.0, min(1.0, h))
        if h != self._h:
            self._h = h
            self._sync_to_rgb()
            self.changed.emit()

    @property
    def s(self) -> float:
        return self._s

    @s.setter
    def s(self, s: float) -> None:
        s = max(0.0, min(1.0, s))
        if s != self._s:
            self._s = s
            self._sync_to_rgb()
            self.changed.emit()

    @property
    def v(self) -> float:
        return self._v

    @v.setter
    def v(self, v: float) -> None:
        v = max(0.0, min(1.0, v))
        if v != self._v:
            self._v = v
            self._sync_to_rgb()
            self.changed.emit()

    @property
    def r(self) -> float:
        return self._r

    @r.setter
    def r(self, r: float) -> None:
        r = max(0.0, min(1.0, r))
        if r != self._r:
            self._r = r
            self._sync_to_hsv()
            self.changed.emit()

    @property
    def g(self) -> float:
        return self._g

    @g.setter
    def g(self, g: float) -> None:
        g = max(0.0, min(1.0, g))
        if g != self._g:
            self._g = g
            self._sync_to_hsv()
            self.changed.emit()

    @property
    def b(self) -> float:
        return self._b

    @b.setter
    def b(self, b: float) -> None:
        b = max(0.0, min(1.0, b))
        if b != self._b:
            self._b = b
            self._sync_to_hsv()
            self.changed.emit()


def black_or_white(color: QtGui.QColor) -> int:
    rgb: T.List[float] = []
    for c in [color.redF(), color.greenF(), color.blueF()]:
        if c <= 0.03928:
            c = c / 12.92
        else:
            c = ((c + 0.055) / 1.055) ** 2.4
        rgb.append(c)
    r, g, b = rgb
    luminance = 0.2126 * r + 0.7152 * g + 0.0722 * b
    if luminance > math.sqrt(1.05 * 0.05) - 0.05:
        return QtCore.Qt.black
    return QtCore.Qt.white



import enum

from PyQt5 import QtCore, QtGui, QtWidgets

from .color_model import ColorModel, black_or_white
from .util import is_imprecise_click, is_precise_click


class ColorSquareStyle(enum.IntEnum):
    Hue = 1
    Saturation = 2
    Value = 3
    Red = 4
    Green = 5
    Blue = 6


class ColorSquare(QtWidgets.QFrame):
    def __init__(
        self,
        parent: QtWidgets.QWidget,
        model: ColorModel,
        square_style: ColorSquareStyle = ColorSquareStyle.Hue,
    ) -> None:
        super().__init__(parent)

        self._square_style = square_style
        self._model = model
        self._pressed = False

        self._model.changed.connect(self.update)

        self.setFrameStyle(QtWidgets.QFrame.Panel | QtWidgets.QFrame.Sunken)
        self.setFixedSize(300, 300)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

    @property
    def square_style(self) -> ColorSquareStyle:
        return self._square_style

    def set_square_style(self, square_style: ColorSquareStyle) -> None:
        self._square_style = square_style
        self.update()

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        rect = self.rect()
        rect -= QtCore.QMargins(*[self.lineWidth()] * 4)

        painter = QtGui.QPainter(self)
        painter.setRenderHint(painter.Antialiasing)

        for y in range(rect.top(), rect.top() + rect.height()):
            x1 = 0
            x2 = self.width()
            gradient = QtGui.QLinearGradient(x1, y, x2, y)
            self._decorate_gradient(gradient, y / rect.height())

            painter.setPen(QtCore.Qt.NoPen)
            painter.setBrush(gradient)
            painter.drawRect(QtCore.QRect(rect.left(), y, rect.width(), 1))

        pos = self._get_color_pos()
        pos.setX(pos.x() * rect.width())
        pos.setY(pos.y() * rect.height())

        cx = pos.x()
        cy = pos.y()

        painter.setPen(QtGui.QPen(black_or_white(self._model.color), 1.5))
        painter.setBrush(QtCore.Qt.NoBrush)
        painter.drawEllipse(QtCore.QRect(cx - 5, cy - 5, 10, 10))

        super().paintEvent(event)

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        is_precise = is_precise_click(event.button(), self.style())
        is_imprecise = is_imprecise_click(event.button(), self.style())

        if is_precise or is_imprecise:
            event.accept()
            self._sync(event.pos())
            if is_imprecise:
                self._pressed = True
            return

        event.ignore()

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        if self._pressed:
            event.accept()
            self._sync(event.pos())
        else:
            event.ignore()

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        self._pressed = None

    def _get_color_pos(self) -> QtCore.QPointF:
        if self._square_style == ColorSquareStyle.Hue:
            return QtCore.QPointF(self._model.s, 1 - self._model.v)
        elif self._square_style == ColorSquareStyle.Saturation:
            return QtCore.QPointF(self._model.h, 1 - self._model.v)
        elif self._square_style == ColorSquareStyle.Value:
            return QtCore.QPointF(self._model.h, 1 - self._model.s)
        elif self._square_style == ColorSquareStyle.Red:
            return QtCore.QPointF(self._model.b, 1 - self._model.g)
        elif self._square_style == ColorSquareStyle.Green:
            return QtCore.QPointF(self._model.b, 1 - self._model.r)
        elif self._square_style == ColorSquareStyle.Blue:
            return QtCore.QPointF(self._model.r, 1 - self._model.g)
        else:
            assert False

    def _decorate_gradient(self, gradient: QtGui.QGradient, y: float) -> None:
        if self._square_style == ColorSquareStyle.Hue:
            for i in [0, 1]:
                gradient.setColorAt(
                    i, QtGui.QColor.fromHsvF(self._model.h, i, 1 - y)
                )

        elif self._square_style == ColorSquareStyle.Saturation:
            for i in range(6):
                gradient.setColorAt(
                    i / 6, QtGui.QColor.fromHsvF(i / 6, self._model.s, 1 - y)
                )

        elif self._square_style == ColorSquareStyle.Value:
            for i in range(6):
                gradient.setColorAt(
                    i / 6, QtGui.QColor.fromHsvF(i / 6, 1 - y, self._model.v)
                )

        elif self._square_style == ColorSquareStyle.Red:
            for i in [0, 1]:
                gradient.setColorAt(
                    i, QtGui.QColor.fromRgbF(self._model.r, 1 - y, i)
                )

        elif self._square_style == ColorSquareStyle.Green:
            for i in [0, 1]:
                gradient.setColorAt(
                    i, QtGui.QColor.fromRgbF(1 - y, self._model.g, i)
                )

        elif self._square_style == ColorSquareStyle.Blue:
            for i in [0, 1]:
                gradient.setColorAt(
                    i, QtGui.QColor.fromRgbF(i, 1 - y, self._model.b)
                )

        else:
            assert False

    def _sync(self, pos: QtCore.QPoint) -> None:
        x = pos.x() / self.rect().width()
        y = pos.y() / self.rect().height()
        if self._square_style == ColorSquareStyle.Hue:
            self._model.s = x
            self._model.v = 1 - y
        elif self._square_style == ColorSquareStyle.Saturation:
            self._model.h = x
            self._model.v = 1 - y
        elif self._square_style == ColorSquareStyle.Value:
            self._model.h = x
            self._model.s = 1 - y
        elif self._square_style == ColorSquareStyle.Red:
            self._model.b = x
            self._model.g = 1 - y
        elif self._square_style == ColorSquareStyle.Green:
            self._model.b = x
            self._model.r = 1 - y
        elif self._square_style == ColorSquareStyle.Blue:
            self._model.r = x
            self._model.g = 1 - y
        else:
            assert False



import math
import sys
import typing as T

from PyQt5 import QtCore, QtGui, QtWidgets

from .color_model import ColorModel, black_or_white
from .geometry import point_in_ring, point_in_triangle
from .util import blend, clamp, is_imprecise_click, is_precise_click


class ColorRing(QtWidgets.QWidget):
    _ring_outer_radius = 150
    _ring_width = 35

    def __init__(self, parent: QtWidgets.QWidget, model: ColorModel) -> None:
        super().__init__(parent)
        self._model = model
        self._model.changed.connect(self.update)

        self._pressed_control: T.Optional[str] = None
        self._ring_gradient = QtGui.QConicalGradient(
            QtCore.QPoint(self._ring_outer_radius, self._ring_outer_radius),
            0.0,
        )
        self._ring_gradient.setColorAt(0 / 6, QtGui.QColor(255, 0, 0, 255))
        self._ring_gradient.setColorAt(1 / 6, QtGui.QColor(255, 0, 255, 255))
        self._ring_gradient.setColorAt(2 / 6, QtGui.QColor(0, 0, 255, 255))
        self._ring_gradient.setColorAt(3 / 6, QtGui.QColor(0, 255, 255, 255))
        self._ring_gradient.setColorAt(4 / 6, QtGui.QColor(0, 255, 0, 255))
        self._ring_gradient.setColorAt(5 / 6, QtGui.QColor(255, 255, 0, 255))
        self._ring_gradient.setColorAt(6 / 6, QtGui.QColor(255, 0, 0, 255))

        self.setFixedSize(self._ring_outer_diameter, self._ring_outer_diameter)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        painter = QtGui.QPainter(self)
        painter.drawImage(0, 0, self._draw_ring())
        painter.drawImage(0, 0, self._draw_triangle())

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        is_precise = is_precise_click(event.button(), self.style())
        is_imprecise = is_imprecise_click(event.button(), self.style())

        if is_precise or is_imprecise:
            if point_in_ring(
                event.pos() - self.rect().center(),
                self._ring_inner_radius,
                self._ring_outer_radius,
            ):
                event.accept()
                self._sync_hue_from_ring(event.pos())
                if is_imprecise:
                    self._pressed_control = "ring"
                return

            if point_in_triangle(event.pos(), self._get_triangle_points()):
                event.accept()
                self._sync_value_and_saturation_from_triangle(event.pos())
                if is_imprecise:
                    self._pressed_control = "triangle"
                return

        event.ignore()

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        if self._pressed_control == "ring":
            event.accept()
            self._sync_hue_from_ring(event.pos())
        elif self._pressed_control == "triangle":
            event.accept()
            self._sync_value_and_saturation_from_triangle(event.pos())
        else:
            event.ignore()

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        self._pressed_control = None

    def _sync_value_and_saturation_from_triangle(
        self, pos: QtCore.QPoint
    ) -> None:
        transform, _ = self._get_triangle_transform().inverted()
        pos = transform.map(pos)

        height = self._triangle_height
        side = self._triangle_side

        p1c, p2c, p3c = self._get_triangle_colors()
        y_ratio = pos.y() / height
        y_ratio = clamp(y_ratio, 0, 1)
        x1 = (1 - y_ratio) / 2
        x2 = 1 - x1
        x_ratio = ((pos.x() / side) - x1) / max(sys.float_info.min, x2 - x1)
        x_ratio = clamp(x_ratio, 0, 1)

        y_c1 = blend(p1c, p2c, y_ratio)
        y_c2 = blend(p1c, p3c, y_ratio)
        x_c = blend(y_c1, y_c2, x_ratio)

        self._model.s = x_c.saturationF()
        self._model.v = x_c.valueF()

    def _sync_hue_from_ring(self, pos: QtCore.QPoint) -> None:
        pos -= self.rect().center()
        x = pos.x()
        y = pos.y()
        theta = (math.atan2(y, x) / (2 * math.pi)) % 1.0
        self._model.h = theta

    @property
    def _ring_inner_radius(self) -> int:
        return self._ring_outer_radius - self._ring_width

    @property
    def _ring_inner_diameter(self) -> int:
        return self._ring_inner_radius * 2

    @property
    def _ring_outer_diameter(self) -> int:
        return self._ring_outer_radius * 2

    @property
    def _triangle_side(self) -> int:
        return int(self._ring_inner_radius * 3 / math.sqrt(3))

    @property
    def _triangle_height(self) -> int:
        return int(self._ring_inner_radius * 3 / 2)

    def _get_triangle_transform(self) -> QtGui.QTransform:
        transform = QtGui.QTransform()
        transform.translate(self._ring_outer_radius, self._ring_outer_radius)
        transform.rotate(self._model.h * 360.0)
        transform.rotate(90.0)
        transform.translate(
            -self._triangle_side / 2, -self._triangle_height * 2 / 3
        )
        return transform

    def _get_triangle_points(
        self, use_transform: bool = True
    ) -> T.Tuple[QtCore.QPoint, QtCore.QPoint, QtCore.QPoint]:
        p1 = QtCore.QPoint(self._triangle_side / 2, 0)
        p2 = QtCore.QPoint(0, self._triangle_height)
        p3 = QtCore.QPoint(self._triangle_side, self._triangle_height)
        if use_transform:
            transform = self._get_triangle_transform()
            p1 = transform.map(p1)
            p2 = transform.map(p2)
            p3 = transform.map(p3)
        return (p1, p2, p3)

    def _get_triangle_colors(
        self
    ) -> T.Tuple[QtGui.QColor, QtGui.QColor, QtGui.QColor]:
        p1c = QtGui.QColor.fromHsvF(self._model.h, 1, 1)
        p2c = QtGui.QColor.fromHsvF(self._model.h, 1, 0)
        p3c = QtGui.QColor.fromHsvF(self._model.h, 0, 1)
        return (p1c, p2c, p3c)

    def _draw_ring(self) -> QtGui.QImage:
        image = QtGui.QImage(
            self._ring_outer_diameter,
            self._ring_outer_diameter,
            QtGui.QImage.Format_ARGB32,
        )
        image.fill(0)

        painter = QtGui.QPainter(image)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)

        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(QtCore.Qt.white)
        painter.drawEllipse(
            0, 0, self._ring_outer_diameter, self._ring_outer_diameter
        )

        painter.setCompositionMode(QtGui.QPainter.CompositionMode_Source)
        painter.setBrush(QtCore.Qt.transparent)
        painter.drawEllipse(
            self._ring_width,
            self._ring_width,
            self._ring_inner_diameter,
            self._ring_inner_diameter,
        )

        painter.setCompositionMode(QtGui.QPainter.CompositionMode_SourceAtop)
        painter.setBrush(QtGui.QBrush(self._ring_gradient))
        painter.drawRect(image.rect())

        theta = self._model.h
        painter.setPen(
            QtGui.QPen(
                black_or_white(QtGui.QColor.fromHsvF(self._model.h, 1, 1)), 1.5
            )
        )
        painter.translate(self._ring_outer_radius, self._ring_outer_radius)
        painter.rotate(self._model.h * 360.0)
        painter.drawLine(0, 0, self._ring_outer_diameter, 0)

        return image

    def _draw_triangle(self) -> QtGui.QImage:
        height = self._triangle_height
        side = self._triangle_side

        image = QtGui.QImage(
            self._ring_outer_diameter,
            self._ring_outer_diameter,
            QtGui.QImage.Format_ARGB32,
        )
        image.fill(0)

        painter = QtGui.QPainter(image)
        painter.setRenderHint(painter.Antialiasing)

        p1, p2, p3 = self._get_triangle_points(use_transform=False)
        p1c, p2c, p3c = self._get_triangle_colors()

        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(QtCore.Qt.black)
        painter.setTransform(self._get_triangle_transform())
        painter.drawPolygon(p1, p2, p3)
        painter.setCompositionMode(QtGui.QPainter.CompositionMode_SourceAtop)

        for y in range(0, height):
            ratio = y / height
            x1 = (1 - ratio) / 2
            x2 = 1 - x1
            c1 = blend(p1c, p2c, ratio)
            c2 = blend(p1c, p3c, ratio)

            gradient = QtGui.QLinearGradient(x1 * side, y, x2 * side, y)
            gradient.setColorAt(0, c1)
            gradient.setColorAt(1, c2)
            painter.setBrush(gradient)
            painter.drawRect(0, y, side, 2)

        v = self._model.v
        s = self._model.s
        cx = math.floor(
            p2.x() + (p3.x() - p2.x()) * v + (p1.x() - p3.x()) * s * v + 0.5
        )
        cy = math.floor(
            p2.y() + (p3.y() - p2.y()) * v + (p1.y() - p3.y()) * s * v + 0.5
        )

        painter.setPen(QtGui.QPen(black_or_white(self._model.color), 1.5))
        painter.setBrush(QtCore.Qt.NoBrush)
        painter.setCompositionMode(QtGui.QPainter.CompositionMode_Source)
        painter.drawEllipse(QtCore.QRect(cx - 5, cy - 5, 10, 10))

        return image



import typing as T

from PyQt5 import QtCore, QtGui, QtWidgets

from .color_model import ColorModel, black_or_white


class ColorPreview(QtWidgets.QFrame):
    def __init__(
        self,
        parent: QtWidgets.QWidget,
        orig_color: QtGui.QColor,
        model: ColorModel,
        alpha_grid: QtGui.QPixmap,
    ) -> None:
        super().__init__(parent)
        self._orig_color = orig_color
        self._model = model
        self._alpha_grid = alpha_grid

        self._model.changed.connect(self.update)

        self.setFrameStyle(QtWidgets.QFrame.Panel | QtWidgets.QFrame.Sunken)

    def sizeHint(self) -> T.Tuple[int, int]:
        return QtCore.QSize(300, 50)

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        rect = self.rect()
        rect -= QtCore.QMargins(*[self.lineWidth()] * 4)

        left = QtCore.QRect(rect)
        left.setWidth(left.width() // 2)
        right = QtCore.QRect(rect)
        right.setLeft(left.width())

        painter = QtGui.QPainter()
        painter.begin(self)
        painter.drawTiledPixmap(rect, self._alpha_grid)
        self._draw_color(painter, left, self._orig_color)
        self._draw_color(painter, right, self._model.color)
        painter.end()

        super().paintEvent(event)

    def _draw_color(
        self, painter: QtGui.QPainter, rect: QtCore.QRect, color: QtGui.QColor
    ) -> None:
        text = f"#{color.red():02X}{color.green():02X}{color.blue():02X}"
        painter.fillRect(rect, color)

        painter.setPen(black_or_white(color))
        painter.drawText(rect, QtCore.Qt.AlignCenter, text)

    def set_color(self, color: QtGui.QColor) -> None:
        self._color = color
        self.update()



#!/usr/bin/env python3
import argparse
import sys
import typing as T

from PyQt5 import QtCore, QtGui, QtWidgets

# from PyQt5.QtWidgets import QColorDialog
from pyqtcolordialog import QColorDialog


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--alpha", action="store_true")
    parser.add_argument("--no-alpha", action="store_false", dest="alpha")
    parser.add_argument("--buttons", action="store_true")
    parser.add_argument("--no-buttons", action="store_false", dest="buttons")
    parser.add_argument("--square", action="store_true")
    parser.add_argument("--no-square", action="store_false", dest="square")
    parser.set_defaults(alpha=True, buttons=True, square=False)
    return parser.parse_args()


def set_bit(
    options: QColorDialog.ColorDialogOptions,
    option: QColorDialog.ColorDialogOption,
    value: bool,
) -> None:
    if value:
        options |= option
    else:
        options &= ~option
    return options


def main() -> None:
    args = parse_args()
    app = QtWidgets.QApplication(sys.argv)

    options = QColorDialog.ColorDialogOptions()
    options = set_bit(options, QColorDialog.ShowAlphaChannel, args.alpha)
    options = set_bit(options, QColorDialog.NoButtons, not args.buttons)
    color = QColorDialog.getColor(
        initial=QtGui.QColor(202, 224, 250),
        title="Fancy title...",
        options=options,
        use_square_view=args.square,
    )

    print(color.isValid())
    print(f"{color.red():02x}{color.green():02x}{color.blue():02x}")


if __name__ == "__main__":
    main()



from setuptools import find_packages, setup

setup(
    author="Marcin Kurczewski",
    author_email="rr-@sakuya.pl",
    name="pyqtcolordialog",
    description=("An improved color dialog for PyQt5"),
    version="0.3",
    url="https://github.com/rr-/pyqtcolordialog",
    packages=find_packages(),
    install_requires=["PyQt5"],
    package_dir={"pyqtcolordialog": "pyqtcolordialog"},
    package_data={"pyqtcolordialog": ["*.png", "../LICENSE.md"]},
    classifiers=[
        "Environment :: X11 Applications :: Qt",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Desktop Environment",
        "Topic :: Software Development",
        "Topic :: Software Development :: Widget Sets",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
